import numpy as np
import torch
import os
import random


class EarlyStopMonitor(object):
    def __init__(self, max_round=3, higher_better=True, tolerance=1e-3):
        self.max_round = max_round
        self.num_round = 0

        self.epoch_count = 0
        self.best_epoch = 0

        self.last_best = None
        self.higher_better = higher_better
        self.tolerance = tolerance

    def early_stop_check(self, curr_val):
        if not self.higher_better:
            curr_val *= -1
        if self.last_best is None:
            self.last_best = curr_val
        elif (curr_val - self.last_best) / np.abs(self.last_best) > self.tolerance:
            self.last_best = curr_val
            self.num_round = 0
            self.best_epoch = self.epoch_count
        else:
            self.num_round += 1
        self.epoch_count += 1
        return self.num_round >= self.max_round


class RandEdgeSampler(object):
    def __init__(self, src_list, dst_list):
        src_list = np.concatenate(src_list)
        dst_list = np.concatenate(dst_list)
        self.src_list = np.unique(src_list)
        self.dst_list = np.unique(dst_list)

    def sample(self, size):
        src_index = np.random.randint(0, len(self.src_list), size)
        dst_index = np.random.randint(0, len(self.dst_list), size)
        return self.src_list[src_index], self.dst_list[dst_index]


def set_random_seed(seed):
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    np.random.seed(seed)
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)


def process_sampling_numbers(num_neighbors, num_layers):
    num_neighbors = [int(n) for n in num_neighbors]
    if len(num_neighbors) == 1:
        num_neighbors = num_neighbors * num_layers
    else:
        num_layers = len(num_neighbors)
    return num_neighbors, num_layers


class RandEdgeSamplerVisited(object):
    def __init__(self, src_list, dst_list, seed=None):
        self.seed = None
        src_list = np.concatenate(src_list)
        dst_list = np.concatenate(dst_list)
        self.src_list = np.unique(src_list)
        self.dst_list = np.unique(dst_list)
        self.sampling_dictionary = {}
        for number in self.src_list:
            index_of_source = np.where(src_list == number)[0]
            destination_values = dst_list[index_of_source]
            destination_values = np.unique(destination_values)
            self.sampling_dictionary[number] = destination_values

        if seed is not None:
            self.seed = seed
            self.random_state = np.random.RandomState(self.seed)

    def sample(self, sources_batch):
        dst_index = []
        src_index = []
        if self.seed is None:
            for src in sources_batch:
                src_index.append(src)
                dst_index.append(np.random.choice(self.sampling_dictionary[src]))
        else:
            try:
                for src in sources_batch:
                    src_index.append(src)
                    dst_index.append(self.random_state.choice(self.sampling_dictionary[src]))
            except TypeError:
                src_index.append(sources_batch)
                dst_index.append(self.random_state.choice(self.sampling_dictionary[sources_batch]))

        return np.array(src_index), np.array(dst_index)

    def reset_random_state(self):
        self.random_state = np.random.RandomState(self.seed)

